from abc import ABC
from abc import abstractmethod

from aio_client.base import RequestTypeEnum
from aio_client.consumer.api import push_request
from aio_client.consumer.models import PostConsumerRequest
from django.db.models import OuterRef
from django.db.models import Subquery
from educommon.ws_log.models import SmevLog
from educommon.ws_log.models import SmevSourceEnum

from concentrator.smev3_v321 import settings
from concentrator.smev3_v321.base.constants import PRIVILEGE_DOC_ISSUED
from concentrator.smev3_v321.base.utils import get_adaptation_program
from concentrator.smev3_v321.base.utils import get_person_identity_doc_info
from concentrator.smev3_v321.base.utils import get_language
from concentrator.smev3_v321.base.utils import get_schedule
from concentrator.smev3_v321.base.utils import get_medical_report_without_files
from concentrator.smev3_v321.base.utils import is_cancel_allowed
from concentrator.smev3_v321.base.utils import render_type2xml
from concentrator.smev3_v321.base.utils import get_address
from concentrator.smev3_v321.base.utils import get_child_info
from concentrator.smev3_v321.base.utils import get_person_info
from concentrator.smev3_v321.models import ApplicantAnswer
from concentrator.smev3_v321.models import OrderRequest
from concentrator.smev3_v321.models import UpdateOrderRequest
from concentrator.smev3_v321.service_types import kinder_order
from kinder.core.declaration.models import Declaration
from kinder.core.declaration.models import DeclarationPrivilege
from kinder.core.declaration_status.models import DeclarationStatusTransfer
from kinder.core.direct.models import DRS
from kinder.core.direct.models import DirectRefusalReason
from kinder.core.direct.models import RefusalReasonInitiatorType

from .constants import DECLARATION_FORMATION
from .constants import DELEGATE_REFUSED
from .constants import DIRECTED_TO_DOO
from .constants import NEED_NOT_CONFIRMED
from .constants import WAITING_FOR_CONTRACT
from .constants import CreateOrderStatusMapper
from .constants import UpdateOrderStatusMapper


class OrderRequestBaseHelper(ABC):
    """Базовый класс отправки запросов OrderRequest."""

    service_name = None

    def __init__(self, declaration_id, order_id, direct=None,
                 delegate_id=None):
        self.declaration = Declaration.objects.get(id=declaration_id)
        self.order_id = order_id
        self.delegate_id = delegate_id

        self.direct = direct or self.declaration.direct_set.all().first()
        self.epgu_status_mapper = CreateOrderStatusMapper(
            self.declaration, self.direct)

    @classmethod
    def log(cls, request_obj, result=None, logging_data=None):
        """
        Логирование запроса СМЭВ 3.

        :param request_obj: Сообщение
        :param result: Результат
        :param logging_data: Дополнительные параметры
        """

        logging_data = logging_data or {}

        data = {
            'service_address':
                RequestTypeEnum.get_url(RequestTypeEnum.CS_POST),
            'method_name': cls.service_name,
            'method_verbose_name':
                f'{cls.service_name} (Формирование и отправка запроса '
                f'OrderRequest "Передача данных заявления в ЛК ЕПГУ".)',
            'direction': SmevLog.OUTGOING,
            'interaction_type': SmevLog.IS_SMEV,
            'source': SmevSourceEnum.CONCENTRATOR,
            'request': getattr(request_obj, 'body', None),
            'result': result,
            **logging_data
        }

        return SmevLog(**data).save()

    @abstractmethod
    def get_request(self):
        """Формирование запроса"""

    @abstractmethod
    def send_request(self):
        """Отправка запроса"""


class OrderRequestHelper(OrderRequestBaseHelper):
    """
    Формирование и отправка запроса OrderRequest
    "Передача данных заявления в ЛК ЕПГУ".

    """

    service_name = 'OrderRequest'

    def get_status_history_data(self):
        """Получение данных по истории статусов заявки"""

        history_data = []

        decl_status_code = self.declaration.status.code

        if not self.declaration.declarationstatuslog_set.exists():

            epgu_status_code, epgu_comment = (
                self.epgu_status_mapper[{'status__code': decl_status_code}]
            )

            return kinder_order.statusHistoryListType(
                [kinder_order.statusHistoryType(
                    statusCode=kinder_order.statusCodeType(
                        orgCode=epgu_status_code),
                    statusDate=self.declaration.date,
                    statusComment=epgu_comment,
                    cancelAllowed=is_cancel_allowed(self.declaration.status))])

        auto_change_days = Subquery(
            DeclarationStatusTransfer.objects.filter(
                from_status=OuterRef('old_status'),
                to_status=OuterRef('status'),
                auto_change_days__gt=0
            ).values('auto_change_days')[:1]
        )

        decl_status_log = self.declaration.declarationstatuslog_set.annotate(
            auto_change_days=auto_change_days
        ).values(
            'id', 'old_status__code', 'status__code', 'datetime',
            'comment', 'auto_change_days'
        ).all()

        for status_history in decl_status_log:
            epgu_status_code, epgu_comment = (
                self.epgu_status_mapper[status_history]
            )

            history_data.append(kinder_order.statusHistoryType(
                statusCode=kinder_order.statusCodeType(
                    orgCode=epgu_status_code
                ),
                statusDate=status_history['datetime'],
                statusComment=epgu_comment,
                cancelAllowed=is_cancel_allowed(self.declaration.status)
            ))

        return kinder_order.statusHistoryListType(history_data)

    def get_edu_organizations_data(self):
        """Получение данных по желаеммым ДОО в заявке"""

        organizations = []

        for decl_unit in self.declaration.declarationunit_set.all():
            organizations.append(kinder_order.EduOrganizationType(
                code=decl_unit.unit_id,
                PriorityNumber=decl_unit.ord,
                valueOf_=decl_unit.unit.name
            ))

        return organizations

    def get_application_data(self):
        """Получение всех данных по заявке"""

        if self.delegate_id:
            delegate = self.declaration.children.childrendelegate_set.get(
                delegate__id=self.delegate_id
            ).delegate
        else:
            delegate = (
                self.declaration.children.childrendelegate_set.first().delegate
            )
        children = self.declaration.children

        application_data = kinder_order.ApplicationType(
            PersonIdentityDocInfo=(
                get_person_identity_doc_info(
                    kinder_order, delegate, self.declaration)),
            PersonInfo=get_person_info(kinder_order, delegate),
            ChildInfo=get_child_info(kinder_order, self.declaration),
            Address=get_address(kinder_order, children),
            EntryParams=kinder_order.EntryParamsType(
                EntryDate=self.declaration.desired_date,
                Language=kinder_order.DataElementType(
                    **get_language(self.declaration.spec)),
                Schedule=kinder_order.DataElementType(
                    **get_schedule(self.declaration.work_type)),
                AgreementOnFullDayGroup=self.declaration.consent_full_time_group
            ),
            AdaptationProgram=(
                get_adaptation_program(kinder_order, self.declaration)),
            MedicalReport=get_medical_report_without_files(
                kinder_order, self.declaration),
            EduOrganizations=kinder_order.EduOrganizationsType(
                EduOrganization=self.get_edu_organizations_data(),
                AllowOfferOther=self.declaration.offer_other))

        for declaration_unit in self.declaration.declarationunit_set.all():
            if declaration_unit.sibling:
                application_data.add_BrotherSisterInfo(
                    kinder_order.BrotherSisterInfoType(
                        ChildSurname=declaration_unit.sibling.surname,
                        ChildName=declaration_unit.sibling.firstname,
                        ChildMiddleName=declaration_unit.sibling.patronymic,
                        EduOrganization=kinder_order.DataElementType(
                            code=declaration_unit.unit_id,
                            valueOf_=declaration_unit.unit.name,
                        ),
                    )
                )

        declaration_privilege = DeclarationPrivilege.objects.filter(
            privilege=self.declaration.best_privilege,
            declaration=self.declaration, privilege__esnsi_code__isnull=False
        ).first()

        if declaration_privilege:

            end_date_info = {}
            if declaration_privilege._privilege_end_date:
                end_date_info['DocExpirationDate'] = (
                    declaration_privilege._privilege_end_date)

            doc_info = kinder_order.DocInfoType(
                DocIssueDate=(
                    declaration_privilege.doc_date or self.declaration.date),
                DocIssued=(
                    declaration_privilege.doc_issued_by
                    or PRIVILEGE_DOC_ISSUED),
                **end_date_info
            )

            application_data.set_BenefitInfo(
                kinder_order.BenefitInfoWithoutFilesType(
                    BenefitCategory=kinder_order.DataElementType(
                        code=declaration_privilege.privilege.esnsi_code,
                        valueOf_=declaration_privilege.privilege.name,
                    ),
                    BenefitDocInfo=doc_info
                )
            )

        return application_data

    def get_request(self):
        request = kinder_order.OrderRequestType(
            env=settings.ORDER_REQUEST_SERVICE_ENV,
            CreateOrderRequest=kinder_order.CreateOrderRequestType(
                orderId_InfoRequest=self.order_id,
                requestDate=self.declaration.date,
                statusHistoryList=self.get_status_history_data(),
                application=self.get_application_data(),)
        )

        request = render_type2xml(
            request,
            name_type=self.service_name
        )

        return request

    def send_request(self):
        request = self.get_request()

        order_request, _ = OrderRequest.objects.get_or_create(
            declaration=self.declaration,
            defaults={'request_order_id': self.order_id}
        )
        order_request.generate_message_id()

        request_obj = PostConsumerRequest(
            message_id=str(order_request.message_id),
            body=request,
            message_type=self.service_name
        )

        push_request(request_obj)

        return request_obj


class UpdateRequestHelper(OrderRequestHelper):
    """
    Формирование и отправка информации об изменении статуса заявления
    по ВС "Передача заявлений на запись в дошкольную организацию".
    Запрос OrderRequest с блоком UpdateOrderRequest.

    """

    service_name = 'UpdateOrderRequest'

    # Именованные аргументы с которыми может вызыватся асинхронная задача и
    # соотвественно данный хелпер. Используются в маппинге статусов для
    # формирования комментария ЕПГУ
    init_kwargs = (
        'decl_status_changed',
        'direct_status_log',
        'event',
        'declaration_changes_rows',
        'reject_changes_comment'
    )

    def __init__(self, declaration_id, order_id, direct=None, **kwargs):
        super().__init__(declaration_id, order_id, direct)

        for kwarg in kwargs:
            assert kwarg in self.init_kwargs

        self.decl_status_changed = kwargs.get('decl_status_changed', False)
        self.direct_status_log = kwargs.get('direct_status_log', None)
        self.event = kwargs.get('event', None)

        self.__dict__.update(kwargs)

        self.epgu_status_mapper = UpdateOrderStatusMapper(
            self.declaration, direct, **kwargs)

    def get_status_history_data(self):
        """
        Получение данных по статусу заявки (код и комментарий для ЕПГУ).

        """

        auto_change_days = Subquery(
            DeclarationStatusTransfer.objects.filter(
                from_status=OuterRef('old_status'),
                to_status=OuterRef('status'),
                auto_change_days__gt=0
            ).values('auto_change_days')[:1]
        )

        current_decl_status_code = self.declaration.status.code
        status_date = self.declaration.date

        # Произошло изменение статуса заявления,
        # достаем текущий установленный статус
        if self.decl_status_changed:
            epgu_status_code, epgu_comment = (
                self.epgu_status_mapper[
                    {'status__code': current_decl_status_code}
                ]
            )
        # Произошло изменение статуса направления
        elif self.direct_status_log:
            new_direct_status = self.direct_status_log['status']

            if new_direct_status == DRS.REGISTER:
                has_applicant_answer = ApplicantAnswer.objects.filter(
                    direct=self.direct
                ).values_list('answer', flat=True).first()
                if has_applicant_answer:
                    epgu_status_code, epgu_comment = (
                        self.epgu_status_mapper[
                            {'status__code': DIRECTED_TO_DOO}
                        ]
                    )
                else:
                    epgu_status_code, epgu_comment = (
                        self.epgu_status_mapper[
                            {'status__code': DECLARATION_FORMATION}
                        ]
                    )
            elif new_direct_status in (DRS.REJECT, DRS.REFUSED):
                initiator_type = DirectRefusalReason.objects.filter(
                    direct_status_log_id=self.direct_status_log['id']
                ).values_list('reason__initiator_type', 'reason__name').first()

                if (initiator_type['reason__initiator_type'] ==
                        RefusalReasonInitiatorType.EMPLOYEE):
                    epgu_status_code, epgu_comment = (
                        self.epgu_status_mapper[
                            {
                                'status__code': NEED_NOT_CONFIRMED,
                                'reason': initiator_type['reason__name'],
                                'created_at': self.direct_status_log['created_at'],
                                'old_status': self.direct_status_log['old_status'],
                            }
                        ]
                    )
                elif initiator_type == RefusalReasonInitiatorType.PARENT:
                    epgu_status_code, epgu_comment = (
                        self.epgu_status_mapper[
                            {'status__code': DELEGATE_REFUSED}
                        ]
                    )
            elif new_direct_status == DRS.DOGOVOR:
                epgu_status_code, epgu_comment = (
                    self.epgu_status_mapper[
                        {'status__code': WAITING_FOR_CONTRACT}
                    ]
                )
        # Произошло событие "Отправка в ЭДС запроса ApplicationRequest" или
        # "В ЭДС пришло изменение и изменения отклонены"
        # (Заявление - вкладка "Изменения с ЕПГУ" - выбрать запись - Отказать)
        elif self.event:
            epgu_status_code, epgu_comment = (
                self.epgu_status_mapper[
                    {'status__code': self.event}
                ]
            )
        # Произошло событие изменение полей заявления (кроме статуса).
        # Достаем последний установленный статус из лога статусов,
        # если он пуст, то берем текуший статус заявки
        else:
            decl_status_log = self.declaration.declarationstatuslog_set.annotate(
                auto_change_days=auto_change_days
            ).values(
                'id', 'old_status__code', 'status__code', 'datetime',
                'comment', 'auto_change_days'
            ).order_by('id').last()

            if decl_status_log:
                epgu_status_code, epgu_comment = (
                    self.epgu_status_mapper[decl_status_log]
                )
                status_date = decl_status_log['datetime']
            else:
                epgu_status_code, epgu_comment = (
                    self.epgu_status_mapper[
                        {'status__code': current_decl_status_code}
                    ]
                )

        return kinder_order.statusHistoryType(
            statusCode=kinder_order.statusCodeType(
                orgCode=epgu_status_code),
            statusDate=status_date,
            statusComment=epgu_comment,
            cancelAllowed=is_cancel_allowed(self.declaration.status)
        )

    def get_request(self):
        """Формирование запроса"""

        if self.decl_status_changed or self.direct_status_log or self.event:
            request = kinder_order.OrderRequestType(
                env=settings.ORDER_REQUEST_SERVICE_ENV,
                UpdateOrderRequest=kinder_order.UpdateOrderRequestType(
                    orderId=int(self.order_id),
                    statusHistoryList=self.get_status_history_data(),
                )
            )
        else:
            request = kinder_order.OrderRequestType(
                env=settings.ORDER_REQUEST_SERVICE_ENV,
                UpdateOrderRequest=kinder_order.UpdateOrderRequestType(
                    orderId=int(self.order_id),
                    statusHistoryList=self.get_status_history_data(),
                    application=self.get_application_data(),
                )
            )

        request = render_type2xml(
            request,
            name_type=self.service_name
        )

        return request

    def send_request(self):
        """Отправка запроса"""

        request = self.get_request()

        update_request = UpdateOrderRequest.objects.create(
            declaration=self.declaration,
        )
        update_request.generate_message_id()

        request_obj = PostConsumerRequest(
            message_id=str(update_request.message_id),
            body=request,
            message_type=self.service_name
        )

        push_request(request_obj)

        return request_obj
